@@ -23,7 +23,8 @@ import inspect
 
 import hamcrest
 
-from .internal import ANY_ARG, OperationList, Method, MockBase, SpyBase, AttributeFactory
+from .internal import (ANY_ARG, OperationList, Method, MockBase, SpyBase,
+                       AttributeFactory)
 from .proxy import create_proxy, get_class
 from .matchers import MockExpectInvocation
 
@@ -35,7 +36,7 @@ __all__ = ['Stub', 'Spy', 'ProxySpy', 'Mock', 'Mimic',
 
 class Stub(object):
     def __new__(cls, collaborator=None):
-        '''Creates a fresh class per instance. This is required due to
+        '''Creates a fresh class clone per instance. This is required due to
         ad-hoc stub properties are class attributes'''
         klass = type(cls.__name__, (cls,), dict(cls.__dict__))
         return object.__new__(klass)
@@ -53,9 +54,8 @@ class Stub(object):
     def __exit__(self, *args):
         self._setting_up = False
 
-    def _manage_invocation(self, invocation, check=True):
-        if check:
-            self._proxy.assure_signature_matches(invocation)
+    def _manage_invocation(self, invocation):
+        self._proxy.assure_signature_matches(invocation)
 
         if self._setting_up:
             self._stubs.append(invocation)
@@ -63,17 +63,15 @@ class Stub(object):
 
         self._prepare_invocation(invocation)
 
-        have_retval = False
+        stubbed_retval = None
         if invocation in self._stubs:
             stubbed = self._stubs.lookup(invocation)
-            retval = stubbed.perform(invocation)
-            have_retval = True
+            stubbed_retval = stubbed.perform(invocation)
 
         actual_retval = self._perform_invocation(invocation)
 
-        if not have_retval:
-            return actual_retval
-
+        retval = stubbed_retval if stubbed_retval is not None else actual_retval
+        invocation.context.retval = retval
         return retval
 
     def _prepare_invocation(self, invocation):
@@ -83,23 +81,30 @@ class Stub(object):
         return None
 
     def __getattr__(self, key):
-        self._add_attr(key)
+        self._add_attr_from_collaborator(key)
         return object.__getattribute__(self, key)
 
-    def _add_attr(self, key):
-        setattr(self.__class__, key, AttributeFactory.create(self, key))
-
     def __setattr__hook(self, key, value):
         if key in self.__dict__:
-            return object.__setattr__(self, key, value)
+            object.__setattr__(self, key, value)
+            return
 
         try:
-            self._add_attr(key)
+            self._add_attr_from_collaborator(key)
         except AttributeError:
+            #  collaborator has not attribute 'key', creaing it ad-hoc
             pass
 
+        # descriptor protocol compliant
         object.__setattr__(self, key, value)
 
+    def _add_attr_from_collaborator(self, key):
+        attr = AttributeFactory.create(self, key)
+        if isinstance(attr, property):
+            setattr(self.__class__, key, attr)
+        else:
+            object.__setattr__(self, key, attr)
+
     def _classname(self):
         name = self._proxy.collaborator_classname()
         return name or self.__class__.__name__
@@ -113,8 +118,9 @@ class Spy(Stub, SpyBase):
     def _prepare_invocation(self, invocation):
         self._recorded.append(invocation)
 
-    def _received_invocation(self, invocation, times):
-        return hamcrest.is_(times).matches(self._recorded.count(invocation))
+    def _received_invocation(self, invocation, times, cmp_pred=None):
+        return hamcrest.is_(times).matches(
+            self._recorded.count(invocation, cmp_pred))
 
     def _get_invocations_to(self, name):
         return [i for i in self._recorded
@@ -131,7 +137,7 @@ class ProxySpy(Spy):
             raise TypeError("ProxySpy takes an instance (got %s instead)" % thing)
 
     def _perform_invocation(self, invocation):
-        return self._proxy.perform_invocation(invocation)
+        return invocation.apply_on_collaborator()
 
 
 class Mock(Spy, MockBase):
