@@ -18,7 +18,6 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-import time
 
 import sys
 from unittest import TestCase
@@ -26,15 +25,37 @@ import itertools
 import thread
 import threading
 
+
 from hamcrest import is_not, all_of, contains_string, has_length
 from hamcrest.library.text.stringcontainsinorder import *
 from hamcrest.library.object.hasproperty import *
 from hamcrest.library.number.ordering_comparison import *
 
 from doublex import *
+from doublex.internal import InvocationContext
 
 
-class StubTests(TestCase):
+class InvocationContextTests(TestCase):
+    def test_order(self):
+        c1 = InvocationContext(1)
+        c2 = InvocationContext(2)
+        contexts = [c2, c1]
+
+        contexts.sort()
+
+        assert_that(contexts[0], is_(c1))
+
+    def test_order_ANY_ARG(self):
+        c1 = InvocationContext(1)
+        c2 = InvocationContext(ANY_ARG)
+        contexts = [c2, c1]
+
+        contexts.sort()
+
+        assert_that(contexts[0], is_(c1))
+
+
+class FreeStubTests(TestCase):
     def setUp(self):
         self.stub = Stub()
 
@@ -85,7 +106,7 @@ class StubTests(TestCase):
             pass
 
 
-class VerifiedStubTests(TestCase):
+class StubTests(TestCase):
     def setUp(self):
         self.stub = Stub(Collaborator)
 
@@ -131,44 +152,59 @@ class VerifiedStubTests(TestCase):
         assert_that(self.stub.kwarg_method(1), is_(1000))
         assert_that(self.stub.kwarg_method(key_param=2), is_(2000))
 
-
-class StubReturnValueTests(TestCase):
-    def setUp(self):
-        self.spy = Spy(Collaborator)
-
     def test_returning_tuple(self):
-        with self.spy:
-            self.spy.hello().returns((3, 4))
+        with self.stub:
+            self.stub.hello().returns((3, 4))
 
-        assert_that(self.spy.hello(), (3, 4))
+        assert_that(self.stub.hello(), (3, 4))
+
+
+class AccessingActualAttributes(TestCase):
+    def test_read_class_attribute_providing_class(self):
+        stub = Stub(Collaborator)
+        assert_that(stub.class_attr, is_("OK"))
+
+    def test_read_class_attribute_providing_instance(self):
+        stub = Stub(Collaborator())
+        assert_that(stub.class_attr, is_("OK"))
+
+    # New in version 1.6.5
+    def test_proxyspy_read_instance_attribute(self):
+        stub = Stub(Collaborator())
+        assert_that(stub.instance_attr, is_(300))
 
 
 class AdhocAttributesTests(TestCase):
     "all doubles accepts ad-hoc attributes"
 
     def test_add_attribute_for_free_stub(self):
-        stub = Stub()
-        stub.foo = 1
+        sut = Stub()
+        sut.foo = 1
 
     def test_add_attribute_for_verified_stub(self):
-        stub = Stub(Collaborator)
-        stub.foo = 1
+        sut = Stub(Collaborator)
+        sut.foo = 1
 
     def test_add_attribute_for_free_spy(self):
-        stub = Spy()
-        stub.foo = 1
+        sut = Spy()
+        sut.foo = 1
 
     def test_add_attribute_for_verified_spy(self):
-        stub = Spy(Collaborator)
-        stub.foo = 1
+        sut = Spy(Collaborator)
+        sut.foo = 1
+
+    def test_add_attribute_for_proxyspy(self):
+        sut = Spy(Collaborator())
+        sut.class_attr = 'other'
 
     def test_add_attribute_for_free_mock(self):
-        stub = Mock()
-        stub.foo = 1
+        sut = Mock()
+        sut.foo = 1
 
     def test_add_attribute_for_verified_mock(self):
-        stub = Mock(Collaborator)
-        stub.foo = 1
+        sut = Mock(Collaborator)
+        sut.foo = 1
+
 
 class SpyTests(TestCase):
     def setUp(self):
@@ -234,13 +270,13 @@ class SpyTests(TestCase):
         self.spy.foo(3.0, [1, 2], 'hi', color='red', width=10)
 
         assert_that(self.spy.foo, called().with_args(
-                3.0, [1, 2], 'hi', color='red', width=10))
+            3.0, [1, 2], 'hi', color='red', width=10))
         assert_that(self.spy.foo, called().with_args(
-                3.0, [1, 2], 'hi', width=10, color='red'))
+            3.0, [1, 2], 'hi', width=10, color='red'))
         assert_that(self.spy.foo, is_not(called().with_args(
-                [1, 2], 'hi', width=10, color='red')))
+            [1, 2], 'hi', width=10, color='red')))
         assert_that(self.spy.foo, is_not(called().with_args(
-                [1, 2], 3.0, 'hi', width=10, color='red')))
+            [1, 2], 3.0, 'hi', width=10, color='red')))
 
     def test_called_with_args_and_times(self):
         self.spy.foo(1)
@@ -267,6 +303,45 @@ class SpyTests(TestCase):
 #        assert_that(spy.mixed_method, called().with_args(key_param=True))
 
 
+class SpyCallsTests(TestCase):
+    def test_list_recorded_calls(self):
+        class Collaborator:
+            def method(self, *args, **kargs):
+                pass
+
+        with Spy(Collaborator) as spy:
+            spy.method(ANY_ARG).returns(100)
+
+        spy.method(1, 2, 3)
+        spy.method(key=2, val=5)
+
+        assert_that(spy.method.calls[0].args, is_((1, 2, 3)))
+        assert_that(spy.method.calls[1].kargs, is_(dict(key=2, val=5)))
+        assert_that(spy.method.calls[1].retval, is_(100))
+
+    def test_called_with_an_object(self):
+        class Module:
+            def getName(self):
+                return "Module"
+
+        class Visitor:
+            def visitModule(self, m):
+                pass
+
+        class Parser:
+            def __init__(self, visitor):
+                self.visitor = visitor
+
+            def accept(self, visitor):
+                self.visitor.visitModule(Module())
+
+        visitor = Spy(Visitor)
+        parser = Parser(visitor)
+        parser.accept(visitor)
+
+        assert_that(visitor.visitModule.calls[0].args[0].getName(), is_("Module"))
+
+
 class VerifiedSpyTests(TestCase):
     def setUp(self):
         self.spy = Spy(Collaborator())
@@ -325,7 +400,7 @@ class BuiltinSpyTests(TestCase):
             assert_that(str(e), contains_string(expected))
 
 
-class ProxySpyTest(TestCase):
+class ProxySpyTests(TestCase):
     def test_must_give_argument(self):
         self.failUnlessRaises(TypeError, ProxySpy)
 
@@ -734,7 +809,7 @@ class StubDelegateTests(TestCase):
         try:
             with self.stub:
                 self.stub.foo().delegates(None)
-            fail("Exception should be raised")
+            self.fail("Exception should be raised")
 
         except WrongApiUsage as e:
             expected = "delegates() must be called with callable or iterable instance (got 'None' instead)"
@@ -828,11 +903,16 @@ class PropertyTests(TestCase):
 
         assert_that(stub.prop, is_(2))
 
-    def test_spy_get_property(self):
+    def test_spy_get_property_using_class(self):
         spy = Spy(ObjCollaborator)
         skip = spy.prop
         assert_that(spy, property_got('prop'))
 
+    def test_spy_get_property_using_instance(self):
+        spy = Spy(ObjCollaborator())
+        skip = spy.prop
+        assert_that(spy, property_got('prop'))
+
     def test_spy_not_get_property(self):
         spy = Spy(ObjCollaborator)
         assert_that(spy, never(property_got('prop')))
@@ -843,11 +923,16 @@ class PropertyTests(TestCase):
             AssertionError,
             assert_that, spy, property_got('prop'))
 
-    def test_spy_set_property(self):
+    def test_spy_set_property_using_class(self):
         spy = Spy(ObjCollaborator)
         spy.prop = 2
         assert_that(spy, property_set('prop'))
 
+    def test_spy_set_property_using_instance(self):
+        spy = Spy(ObjCollaborator())
+        spy.prop = 2
+        assert_that(spy, property_set('prop'))
+
     def test_spy_not_set_property(self):
         spy = Spy(ObjCollaborator)
         assert_that(spy, never(property_set('prop')))
@@ -903,7 +988,28 @@ class PropertyTests(TestCase):
         spy = Spy(ObjCollaborator)
         spy.prop = 2
         spy.prop = 3
-        assert_that(spy, property_set('prop').to(greater_than(1)).times(less_than(3)))
+        assert_that(spy,
+                    property_set('prop').to(greater_than(1)).
+                    times(less_than(3)))
+
+    def test_proxy_spy_get_actual_property(self):
+        collaborator = ObjCollaborator()
+        print collaborator.prop
+        sut = ProxySpy(collaborator)
+        assert_that(sut.prop, is_(1))
+
+    def test_proxy_spy_get_stubbed_property(self):
+        collaborator = ObjCollaborator()
+        with ProxySpy(collaborator) as sut:
+            sut.prop = 2
+        assert_that(sut.prop, is_(2))
+
+    def test_proxy_spy_set_property(self):
+        collaborator = ObjCollaborator()
+        sut = ProxySpy(collaborator)
+        sut.prop = 20
+        assert_that(sut.prop, is_(20))
+        assert_that(collaborator.prop, is_(20))
 
 
 class AsyncTests(TestCase):
@@ -924,7 +1030,7 @@ class AsyncTests(TestCase):
 
         # when
         sut.send_data()
-        barrier.wait(1)    # test probably FAILS without this
+        barrier.wait(1)  # test probably FAILS without this
 
         # then
         assert_that(spy.write, called())
@@ -940,10 +1046,12 @@ class AsyncTests(TestCase):
         # then
         assert_that(spy.write, called().async(timeout=1))
 
-    def test_spy_async_call_fluent_methods(self):
+    def test_spy_async_support_1_call_only(self):
+        # given
         spy = Spy()
         sut = AsyncTests.SUT(spy)
 
+        # when
         sut.send_data(3)
         sut.send_data(3)
 
@@ -951,6 +1059,19 @@ class AsyncTests(TestCase):
         with self.assertRaises(WrongApiUsage):
             assert_that(spy.write, called().async(timeout=1).with_args(3).times(2))
 
+    def test_spy_async_stubbed(self):
+        # given
+        with Spy() as spy:
+            spy.write(ANY_ARG).returns(100)
+
+        sut = AsyncTests.SUT(spy)
+
+        # when
+        sut.send_data(3)
+
+        # then
+        assert_that(spy.write, called().async(timeout=1))
+
 
 class Observer(object):
     def __init__(self):
@@ -980,15 +1101,14 @@ class ObjCollaborator(object):
         return 2
 
 
-#----------------------------#
-#- pyDoubles migrated tests -#
-#----------------------------#
-
 class Collaborator:
     """
     The original object we double in tests
     """
-    test_field = "OK"
+    class_attr = "OK"
+
+    def __init__(self):
+        self.instance_attr = 300
 
     def hello(self):
         return "hello"
